# !/usr/bin/python3
"""main powermon code"""

import json
import logging
import time
from argparse import ArgumentParser
from datetime import date, timedelta  # noqa: F401
from platform import python_version

import yaml
from pydantic import ValidationError

from powermon.version import __version__  # noqa: F401
from powermon.commands.command import Command
from powermon.config.config_model import ConfigModel
from powermon.device import Device
from powermon.libs.apicoordinator import ApiCoordinator
from powermon.libs.daemon import Daemon
from powermon.libs.mqttbroker import MqttBroker
from powermon.protocols import list_protocols

# Set-up logger
log = logging.getLogger("")


def read_yaml_file(yaml_file=None):
    """function to read a yaml file and return dict"""
    _yaml = {}
    if yaml_file is not None:
        try:
            with open(yaml_file, "r", encoding="utf-8") as stream:
                _yaml = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            log.error("Error processing yaml file: %s", exc)
        except FileNotFoundError as exc:
            log.error("Error opening yaml file: %s", exc)
    return _yaml


def process_command_line_overrides(args):
    """override config with command line options"""
    _config = {}
    if args.config:
        _config = json.loads(args.config)
    if args.once:
        _config["loop"] = "once"
    if args.info:
        _config["debuglevel"] = logging.INFO
    if args.debug:
        _config["debuglevel"] = logging.DEBUG
    return _config


def main():
    """main entry point for powermon command"""
    description = f"Power Device Monitoring Utility, version: {__version__}, python version: {python_version()}"  # pylint: disable=C0301
    parser = ArgumentParser(description=description)

    parser.add_argument(
        "-C",
        "--configFile",
        nargs="?",
        type=str,
        help="Full location of config file",
        const="./powermon.yaml",
        default=None,
    )
    parser.add_argument(
        "--config",
        type=str,
        default=None,
        help="""Supply config items on the commandline in json format, \
             eg '{"device": {"port":{"type":"test"}}, "commands": [{"command":"QPI"}]}'""",
    )
    parser.add_argument("-V", "--validate", action="store_true", help="Validate the configuration")
    parser.add_argument("-v", "--version", action="store_true", help="Display the version")
    parser.add_argument("--listProtocols", action="store_true", help="Display the currently supported protocols")
    parser.add_argument(
        "-1",
        "--once",
        action="store_true",
        help="Only loop through config once",
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Force commands to run even if wouldnt be triggered (should only be used with --once)",
    )
    parser.add_argument(
        "-D",
        "--debug",
        action="store_true",
        help="Enable Debug and above (i.e. all) messages",
    )
    parser.add_argument("-I", "--info", action="store_true", help="Enable Info and above level messages")
    parser.add_argument(
        "--adhoc",
        type=str,
        default=None,
        help="Send adhoc command to mqtt adhoc command queue - needs config file specified and populated",
    )

    args = parser.parse_args()
    # prog_name = parser.prog

    # Temporarily set debug level based on command line options
    log.setLevel(logging.WARNING)
    if args.info:
        log.setLevel(logging.INFO)
    if args.debug:
        log.setLevel(logging.DEBUG)

    # Display version if asked
    log.info(description)
    if args.version:
        print(description)
        return None

    # Do enquiry commands
    # - List Protocols
    if args.listProtocols:
        list_protocols()
        return None

    # Build configuration from config file and command line overrides
    log.info("Using config file: %s", args.configFile)
    # build config with details from config file
    config = read_yaml_file(args.configFile)

    # build config - override with any command line arguments
    config.update(process_command_line_overrides(args))

    # validate config
    try:
        config_model = ConfigModel(config=config)
        log.debug(config_model)
        log.info("Config validation successful")
        if args.validate:
            # if --validate option set, only do validation
            print("Config validation successful")
            return None
    except ValidationError as exception:
        # if config fails to validate, print reason and exit
        print("Config validation failed")
        print(f"{config=}")
        print(exception)
        return None

    # logging (DEBUG, INFO, WARNING, ERROR, CRITICAL)
    log.setLevel(config.get("debuglevel", logging.WARNING))

    # debug config
    log.info("config: %s", config)

    # build mqtt broker object (optional)
    mqtt_broker = MqttBroker.from_config(config=config.get("mqttbroker"))
    log.info(mqtt_broker)

    # build device object (required)
    device = Device.from_config(config=config.get("device"))
    device.set_mqtt_broker(mqtt_broker)
    log.debug(device)
    # add commands to device command list
    for command_config in config.get("commands"):
        device.add_command(Command.from_config(command_config))
    log.info(device)

    # build the daemon object (optional)
    daemon = Daemon.from_config(config=config.get("daemon"))
    log.info(daemon)

    # build api coordinator
    api_coordinator = ApiCoordinator.from_config(config=config.get("api"))
    api_coordinator.set_device(device)
    api_coordinator.set_mqtt_broker(mqtt_broker)
    log.info(api_coordinator)

    # initialize api coordinator
    api_coordinator.initialize()

    # initialize daemon
    daemon.initialize()
    api_coordinator.announce(daemon)

    # initialize device
    device.initialize()
    api_coordinator.announce(device)

    # Main working loop
    try:
        while True:
            # tell the daemon we're still working
            daemon.watchdog()

            # run device loop (ie run any needed commands)
            device.run(args.force)

            # run api coordinator ...
            api_coordinator.run()

            # only run loop once if required
            if config.get("loop") == "once":
                break

            # add small delay in loop
            time.sleep(0.1)

    except KeyboardInterrupt:
        print("KeyboardInterrupt - stopping")
    finally:
        # disconnect device
        device.finalize()

        # disconnect mqtt
        mqtt_broker.stop()

        # stop the daemon
        daemon.stop()



""" protocols / abstractprotocol.py """
import abc
import logging
import re

from mppsolar.protocols.protocol_helpers import crcPI as crc
# from powermon.commands.command import Command
from powermon.commands.command_definition import CommandDefinition
from powermon.commands.reading_definition import ReadingDefinition
# from powermon.commands.result import Result, ResultType
from powermon.dto.command_definition_dto import CommandDefinitionDTO
from powermon.dto.protocolDTO import ProtocolDTO
from powermon.errors import PowermonProtocolError, PowermonWIP

log = logging.getLogger("AbstractProtocol")


class AbstractProtocol(metaclass=abc.ABCMeta):
    """ base definition for all protocols """

    def __init__(self) -> None:
        self._command = None
        self._command_dict = None
        self.command_definitions: dict[str, CommandDefinition] = {}
        self.STATUS_COMMANDS = None
        self.SETTINGS_COMMANDS = None
        self.DEFAULT_COMMAND = None
        self.ID_COMMANDS = None
        self._protocol_id = None

    def check_definitions_count(self):
        """ check and report number of command definitions, error if 0 """
        definitions_count = len(self.command_definitions)
        if definitions_count == 0:
            raise PowermonProtocolError(f"Attempted to load protocol '{self._protocol_id}' which has no valid commands")
        log.info("Using protocol:%s with %i commands", self._protocol_id, definitions_count)
        # log.info(f'Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands')

    def to_dto(self) -> ProtocolDTO:
        """ convert protocol object to data transfer object """
        dto = ProtocolDTO(protocol_id=self._protocol_id, commands=self.get_command_definition_dtos())
        return dto

    def get_command_definition_dtos(self) -> dict[str, CommandDefinitionDTO]:
        """ convert all associated command objects to data transfer objects """
        command_dtos: dict[str, CommandDefinitionDTO] = {}
        for command_tuple in self.command_definitions.items():
            command_dtos[command_tuple[0]] = command_tuple[1].to_dto()
        return command_dtos

    def add_command_definitions(self, command_definitions_config: dict):
        """ Add command definitions from the configuration """
        for command_definition_key in command_definitions_config.keys():
            try:
                log.debug("Attempting to add command_definition_key: %s", command_definition_key)
                command_definition = CommandDefinition.from_config(command_definitions_config[command_definition_key])
                self.command_definitions[command_definition_key] = command_definition
            except ValueError as value_error:
                log.info("couldnt add command definition for code: %s", command_definition_key)
                log.info("error was: %s", value_error)

    def list_commands(self) -> dict[str, CommandDefinition]:
        """ list available commands for the protocol """
        if self._protocol_id is None:
            log.error("Attempted to list commands with no protocol defined")
            raise ValueError("Attempted to list commands with no protocol defined")
        return self.command_definitions

    def get_protocol_id(self) -> bytes:
        """ return the protocol id """
        return self._protocol_id

    def get_full_command(self, command) -> bytes:
        """ generate the full command including crc and \n as needed """
        log.info("Using protocol: %s with %i commands", self._protocol_id, len(self.command_definitions))
        byte_cmd = bytes(command, "utf-8")
        # calculate the CRC
        crc_high, crc_low = crc(byte_cmd)
        # combine byte_cmd, CRC , return
        full_command = byte_cmd + bytes([crc_high, crc_low, 13])
        log.debug("full command: %s", full_command)
        return full_command

    def get_response_definition(self, command_definition: CommandDefinition, index=None, key=None) -> ReadingDefinition:
        """ get the definition of a specific response component """
        # QUESTION: is this not a readingdefinition? ie get_reading_definition??
        definitions_count = command_definition.get_response_definition_count()
        if index is not None:
            if index < definitions_count:
                return command_definition.reading_definitions[index]
            else:
                # return [index, f"Unknown value in response {index}", "bytes.decode", ""]
                raise IndexError(f"Index {index} out of range for command {command_definition.code}")
        elif key is not None:
            log.error("key todo abprotocol line 80")  # TODO: add key type get response defn
            raise PowermonWIP("get_response_defn needs key logic implemented")
        else:
            raise PowermonWIP("get_response_defn needs index or key")

    def get_command_with_command_string(self, command) -> CommandDefinition:
        """
        Get the command definition for a given command string
        """

        # Handle the commands that don't have a regex
        if command in self.command_definitions and self.command_definitions[command].regex is None:
            log.debug("Found command %s in protocol %s", command, self._protocol_id)
            return self.command_definitions[command]

        # Try the regex commands
        for command_code, command_definition in self.command_definitions.items():
            if command_definition.regex is not None:
                log.debug("Regex commands _command: %s", command_code)
                _re = re.compile(command_definition.regex)
                match = _re.match(command)
                if match:
                    log.debug("Matched: %s to: %s value: %s", command, command_definition.code, match.group(1))
                    # QUESTION: Is this the only spot to set a parameter for a command?
                    command_definition.set_parameter_value(match.group(1))
                    return command_definition
        log.info("No command_defn found for %s", command)
        return None

    def check_crc(self, response: str):
        """ crc check, needs override in protocol """
        log.debug("check crc for %s", response)
        return

    def check_response_and_trim(self, response: str) -> str:
        """
        Simplest validity check, CRC checks should be added to individual protocols
        """
        log.debug("response: %s", response)
        if response is None:
            raise ValueError("Response is None")
        if len(response) <= 3:
            raise ValueError("Response is too short")
        self.check_crc(response)
        response = response[1:-3]
        return response



import logging
from .abstractprotocol import AbstractProtocol
from .protocol_helpers import crcJK232 as crc

log = logging.getLogger("jk232")


COMMANDS = {
    "getBalancerData": {
        "name": "getBalancerData",
        "command_code": "00",
        "description": "Get Balancer Data",
        "help": " -- Get Balancer Data",
        "type": "QUERY",
        "checksum_required": "True",
        "response_type": "POSITIONAL",
        "response": [
            ["discard", 11, "Packet header", ""],
            ["discard", 1, "Cells connected", ""],
            ["Hex2Int:r/3", 1, "Cells_connected", ""],
            ["discard", 1, "Voltage_Cell01", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell01", "V"],
            ["discard", 1, "Voltage_Cell02", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell02", "V"],
            ["discard", 1, "Voltage_Cell03", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell03", "V"],
            ["discard", 1, "Voltage_Cell04", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell04", "V"],
            ["discard", 1, "Voltage_Cell05", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell05", "V"],
            ["discard", 1, "Voltage_Cell06", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell06", "V"],
            ["discard", 1, "Voltage_Cell07", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell07", "V"],
            ["discard", 1, "Voltage_Cell08", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell08", "V"],
            ["discard", 1, "Voltage_Cell09", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell09", "V"],
            ["discard", 1, "Voltage_Cell10", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell10", "V"],
            ["discard", 1, "Voltage_Cell11", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell11", "V"],
            ["discard", 1, "Voltage_Cell12", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell12", "V"],
            ["discard", 1, "Voltage_Cell13", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell13", "V"],
            ["discard", 1, "Voltage_Cell14", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell14", "V"],
            ["discard", 1, "Voltage_Cell15", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell15", "V"],
            ["discard", 1, "Voltage_Cell16", ""],
            ["BigHex2Short:r/1000", 2, "Voltage_Cell16", "V"],
            ["discard", 1, "MOS_Temp", ""],
            ["BigHex2Short", 2, "MOS_Temp", "°C"],
            ["discard", 1, "Battery_T1", ""],
            ["BigHex2Short", 2, "Battery_T1", "°C"],
            ["discard", 1, "Battery_T2", ""],
            ["BigHex2Short", 2, "Battery_T2", "°C"],
            ["discard", 1, "Battery_Voltage", ""],
            ["BigHex2Short:r/100", 2, "Battery_Voltage", "V"],
            ["discard", 1, "Battery_Current", ""],
            ["BigHex2Short:(r&0x7FFF)/100*(((r&0x8000)>>15)*2-1)", 2, "Battery_Current", "A"],
            ["discard", 1, "Percent_Remain", ""],
            ["Hex2Int", 1, "Percent_Remain", "%"],
            ["discard", 2, "Number of battery sensors", ""],  # useless so dropped
            ["discard", 1, "Cycle_Count", ""],
            ["BigHex2Short", 2, "Cycle_Count", ""],
            ["discard", 1, "Total_capacity", ""],
            ["BigHex2Float", 4, "Total_capacity", "Ahr"], # Needs other formula
            ["discard", 3, "Total number of battery strings", ""],  # dropping
            ["discard", 1, "Battery Warning Message", ""],
            ["Hex2Str", 2, "Battery Warning Message", ""],
            ["discard", 1, "Battery status information", ""],
            ["Hex2Str", 2, "Battery status information", ""],
            ["discard", 15 * 3, "settings", ""],
            ["discard", 1, "Balancer Active", ""],
            ["Hex2Int", 1, "Balancer Active", ""],
            ["discard", 7 * 3, "more settings", ""],
            ["discard", 4 * 3, "temp settings", ""],
            ["discard", 2, "string count", ""], # dropping
            ["discard", 1, "Capacity Setting", ""],
            ["BigHex2Float", 4, "Capacity Setting", "Ahr"],
            ["discard", 1, "Charge Enabled", ""],
            ["Hex2Int", 1, "Charge Enabled", ""],
            ["discard", 1, "Discharge Enabled", ""],
            ["Hex2Int", 1, "Discharge Enabled", ""],
            ["discard", 20 + 96, "remaining data", ""]
        ],
        "test_responses": [
            bytes.fromhex(
                "4e 57 01 1b 00 00 00 00 03 00 01 79 2a 01 0f 91 02 0f 94 03 0f 99 04 0f 92 05 0f 94 06 0f 94 07 0f 94 08 0f 91 09 0f 96 0a 0f 91 0b 0f 92 0c 0f 93 0d"
            ),
        ],
    },
}


class jkserial(AbstractProtocol):
    def __str__(self):
        return "JKBMS TTL serial communication protocol handler"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"JKSERIAL"
        self.COMMANDS = COMMANDS
        self.STATUS_COMMANDS = [
            "getBalancerData",
        ]
        self.SETTINGS_COMMANDS = [
            "",
        ]
        self.DEFAULT_COMMAND = "getBalancerData"

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different
        """
        log.debug(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        # End of required variables setting
        if self._command_defn is None:
            # Maybe return a default here?
            return None
        if "command_code" in self._command_defn:

            # Read basic information and status
            # full command is 21 bytes long
            cmd = bytearray(21)
            command_code = int(self._command_defn["command_code"], 16)

            # start bit  0x4E
            cmd[0] = 0x4E                         # start sequence
            cmd[1] = 0x57                         # start sequence
            cmd[2] = 0x00                         # data length lb
            cmd[3] = 0x13                         # data length hb
            cmd[4] = 0x00                         # bms terminal number
            cmd[5] = 0x00                         # bms terminal number
            cmd[6] = 0x00                         # bms terminal number
            cmd[7] = 0x00                         # bms terminal number
            if self._command_defn["type"] == "SETTER":
                cmd[8] = 0x02                     # command word: 0x01 (activation), 0x02 (write), 0x03 (read), 0x05 (password), 0x06 (read all)
            else:
                cmd[8] = 0x03
            cmd[9] = 0x03                         # frame source: 0x00 (bms), 0x01 (bluetooth), 0x02 (gps), 0x03 (computer)
            cmd[10] = 0x00                        # frame type: 0x00 (read data), 0x01 (reply frame), 0x02 (BMS active upload)
            cmd[11] = command_code                # register: 0x00 (read all registers), 0x8E...0xBF (holding registers)
            cmd[12] = 0x00                        # record number
            cmd[13] = 0x00                        # record number
            cmd[14] = 0x00                        # record number
            cmd[15] = 0x00                        # record number
            cmd[16] = 0x68                        # end sequence
            cmd[17] = 0x00                        # crc unused
            cmd[18] = 0x00                        # crc unused
            crc_high, crc_low = crc(cmd[0:17])
            cmd[19] = crc_high
            cmd[20] = crc_low

            log.debug(f"cmd with crc: {cmd}")
            return cmd

    def get_responses(self, response):
        """
        Override the default get_responses as its different
        """
        responses = []
        # remove \n
        # response = response.replace(b"\n", b"")

        if self._command_defn is not None and self._command_defn["response_type"] == "POSITIONAL":
            # Have a POSITIONAL type response, so need to break it up...
            # example defn :
            # "response": [
            #   ["discard", 1, "start flag", ""],
            #   ["discard", 1, "module address", ""],
            #   ["discard", 1, "command id", ""],
            #   ["discard", 1, "data length", ""],
            # ]
            # example response data  b'NW\x01\x1b\x00\x00\x00\x00\x03\x00\x01y*\x01\x0f\x91\x02\x0f\x94\x03\x0f\x97\x04\x0f\x91\x05\x0f\x94\x06\x0f\x94\x07\x0f\x93\x08\x0f\x92\t\x0f\x96\n\x0f\x91\x0b\x0f\x92\x0c\x0f\x92\r'
            for defn in self._command_defn["response"]:
                size = defn[1]
                item = response[:size]
                responses.append(item)
                response = response[size:]
            if response:
                responses.append(response)
            log.debug(f"get_responses: responses {responses}")
            return responses
        else:
            return bytearray(response)



import logging

from .pi17 import pi17


log = logging.getLogger("pi17infini")

QUERY_COMMANDS = {}
SETTER_COMMANDS = {}
COMMANDS_TO_REMOVE = []


class pi17infini(pi17):
    def __str__(self):
        return "PI17INFINI protocol handler"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI17INFINI"
        # Add pi17infini specific commands to pi17 commands
        self.COMMANDS.update(QUERY_COMMANDS)
        # Add pi17infini specific setter commands
        self.COMMANDS.update(SETTER_COMMANDS)
        # remove and unwanted pi30 commands
        for item in COMMANDS_TO_REMOVE:
            if item in self.COMMANDS:
                self.COMMANDS.pop(item)



import logging
#import datetime.datetime

from .pi30max import pi30max

log = logging.getLogger("pi30m044")

QUERY_COMMANDS = {
    "QPIRI": {
        "name": "QPIRI",
        "description": "Current Settings inquiry",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [
                1,
                "Grid Rating Voltage",
                "float",
                "V",
                {"icon": "mdi:transmission-tower-import", "device-class": "voltage"},
            ],
            [2, "Grid Rating Current", "float", "A", {"icon": "mdi:current-ac", "device-class": "current"}],
            [
                3,
                "AC Output Rating Voltage",
                "float",
                "V",
                {"icon": "mdi:transmission-tower-export", "device-class": "voltage"},
            ],
            [4, "AC Output Rating Frequency", "float", "Hz", {"icon": "mdi:current-ac", "device-class": "frequency"}],
            [5, "AC Output Rating Current", "float", "A", {"icon": "mdi:current-ac", "device-class": "current"}],
            [6, "AC Output Rating Apparent Power", "int", "VA", {"icon": "mdi:power-plug", "device-class": "apparent_power"}],
            [
                7,
                "AC Output Rating Active Power",
                "int",
                "W",
                {"icon": "mdi:power-plug", "device-class": "power", "state_class": "measurement"},
            ],
            [8, "Battery Rating Voltage", "float", "V", {"icon": "mdi:battery-outline", "device-class": "voltage"}],
            [9, "Battery Recharge Voltage", "float", "V", {"icon": "mdi:battery-outline", "device-class": "voltage"}],
            [10, "Battery Under Voltage", "float", "V", {"icon": "mdi:battery-outline", "device-class": "voltage"}],
            [
                11,
                "Battery Bulk Charge Voltage",
                "float",
                "V",
                {"icon": "mdi:battery-outline", "device-class": "voltage"},
            ],
            [
                12,
                "Battery Float Charge Voltage",
                "float",
                "V",
                {"icon": "mdi:battery-outline", "device-class": "voltage"},
            ],
            [
                13,
                "Battery Type",
                "option",
                [
                    "AGM",
                    "Flooded",
                    "User",
                    "Pylontech",
                    "Shinheung",
                    "WECO",
                    "Soltaro",
                    "TBD",
                    "LIb-protocol compatible",
                    "3rd party Lithium",
                ],
            ],
            [14, "Max AC Charging Current", "int", "A", {"icon": "mdi:current-ac", "device-class": "current"}],
            [15, "Max Charging Current", "int", "A", {"icon": "mdi:current-ac", "device-class": "current"}],
            [16, "Input Voltage Range", "option", ["Appliance", "UPS"]],
            [
                17,
                "Output Source Priority",
                "option",
                [
                    "Utility Solar Battery",
                    "Solar Utility Battery",
                    "Solar Battery Utility",
                ],
            ],
            [
                18,
                "Charger Source Priority",
                "option",
                [
                    "Undefined",
                    "Solar first",
                    "Solar + Utility",
                    "Only solar charging permitted",
                ],
            ],
            [19, "Max Parallel Units", "int", "units"],
            [
                20,
                "Machine Type",
                "str_keyed",
                {"00": "Grid tie", "01": "Off Grid", "10": "Hybrid"},
            ],
            [21, "Topology", "option", ["transformerless", "transformer"]],
            [
                22,
                "Output Mode",
                "option",
                [
                    "single machine output",
                    "parallel output",
                    "Phase 1 of 3 Phase output",
                    "Phase 2 of 3 Phase output",
                    "Phase 3 of 3 Phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output (120°)",
                    "Phase 2 of 2 phase output (180°)",
                    "unknown output",
                ],
            ],
            [23, "Battery Redischarge Voltage", "float", "V"],
            [
                24,
                "PV OK Condition",
                "option",
                [
                    "As long as one unit of inverters has connect PV, parallel system will consider PV OK",
                    "Only All of inverters have connect PV, parallel system will consider PV OK",
                ],
            ],
            [
                25,
                "PV Power Balance",
                "option",
                [
                    "PV input max current will be the max charged current",
                    "PV input max power will be the sum of the max charged power and loads power",
                ],
            ],
            [26, "Max charging time for CV stage", "int", "min"],
            [
                27,
                "Operation Logic",
                "option",
                ["Automatic mode", "On-line mode", "ECO mode"],
            ],
            [28, "Max discharging current", "int", "A", {"icon": "mdi:current-ac", "device-class": "current"}],
        ],
        "test_responses": [
            b"(230.0 21.7 230.0 50.0 21.7 5000 4000 48.0 46.0 42.0 56.4 54.0 0 10 010 1 0 0 6 01 0 0 54.0 0 1\x6F\x7E\r",
        ],
    },
    "QFLAG": {
        "name": "QFLAG",
        "description": "Flag Status inquiry",
        "help": " -- queries the enabled / disabled state of various Inverter settings (e.g. buzzer, overload, interrupt alarm)",
        "type": "QUERY",
        "response": [
            [
                "enflags",
                "Device Status",
                {
                    "a": {"name": "Buzzer", "state": "disabled"},
                    "b": {"name": "Overload Bypass", "state": "disabled"},
                    "j": {"name": "Power Saving", "state": "disabled"},
                    "k": {"name": "LCD Reset to Default", "state": "disabled"},
                    "u": {"name": "Overload Restart", "state": "disabled"},
                    "v": {"name": "Over Temperature Restart", "state": "disabled"},
                    "x": {"name": "LCD Backlight", "state": "disabled"},
                    "y": {
                        "name": "Primary Source Interrupt Alarm",
                        "state": "disabled",
                    },
                    "z": {"name": "Record Fault Code", "state": "disabled"},
                },
            ]
        ],
        "test_responses": [
            b"(EbuvyzDajkxn@\r",
        ],
    },
    "QMUCHGCR": {
        "name": "QMUCHGCR",
        "description": "Max Utility Charging Current Options inquiry",
        "help": " -- queries the maximum utility charging current setting of the Inverter",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            ["int", "Max Utility Charging Current 1", "A"],
            ["int", "Max Utility Charging Current 2", "A"],
            ["int", "Max Utility Charging Current 3", "A"],
            ["int", "Max Utility Charging Current 4", "A"],
            ["int", "Max Utility Charging Current 5", "A"],
            ["int", "Max Utility Charging Current 6", "A"],
            ["int", "Max Utility Charging Current 7", "A"],
            ["int", "Max Utility Charging Current 8", "A"],
            ["int", "Max Utility Charging Current 9", "A"],
            ["int", "Max Utility Charging Current 10", "A"],
            ["int", "Max Utility Charging Current 11", "A"],
            ["int", "Max Utility Charging Current 12", "A"],
            ["int", "Max Utility Charging Current 13", "A"],
        ],
        "test_responses": [
            b"(002 010 020 030 040 050 060 070 080 090 100 110 120\xca#\r",
        ],
    },
    "QMCHGCR": {
        "name": "QMCHGCR",
        "description": "Max Charging Current Options inquiry",
        "help": " -- queries the maximum charging current setting of the Inverter",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            ["int", "Max Charging Current 1", "A"],
            ["int", "Max Charging Current 2", "A"],
            ["int", "Max Charging Current 3", "A"],
            ["int", "Max Charging Current 4", "A"],
            ["int", "Max Charging Current 5", "A"],
            ["int", "Max Charging Current 6", "A"],
            ["int", "Max Charging Current 7", "A"],
            ["int", "Max Charging Current 8", "A"],
            ["int", "Max Charging Current 9", "A"],
            ["int", "Max Charging Current 10", "A"],
            ["int", "Max Charging Current 11", "A"],
            ["int", "Max Charging Current 12", "A"],
            ["int", "Max Charging Current 13", "A"],
        ],
        "test_responses": [
            b"(010 020 030 040 050 060 070 080 090 100 110 120\x0c\xcb\r",
        ],
    },
    "QOPPT": {
        "name": "QOPPT",
        "description": "Device Output Source Priority Time Order Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [
            [
                "option",
                "Output Source Priority 00 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 01 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 02 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 03 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 04 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 05 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 06 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 07 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 08 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 09 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 10 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 11 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 12 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 13 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 14 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 15 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 16 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 17 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 18 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 19 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 20 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 21 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 22 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Output Source Priority 23 hours",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Device Output Source Priority",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Selection of Output Source Priority Order 1",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Selection of Output Source Priority Order 2",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
            [
                "option",
                "Selection of Output Source Priority Order 3",
                ["Utility Solar Battery (USB)", "Solar Utility Battery (SUB)", "Solar Battery Utility (SBU)"],
            ],
        ],
        "test_responses": [
            b"(2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 0 2 1>>\r",
        ],
    },
    "QT": {
        "name": "QT",
        "description": "Device Time Inquiry",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "response": [["bytes.decode:datetime.strptime(r, '%Y%m%d%H%M%S').strftime('%Y-%m-%d %H:%M:%S')", "Device Time", ""]],
        "test_responses": [
            b"(20210726122606JF\r",
        ],
    },
    "QBMS": {
        "name": "QBMS",
        "description": "Read lithium battery information",
        "help": " -- queries the value of various metrics from the battery",
        "type": "QUERY",
        "response_type": "SEQUENTIAL",
        "crctype": "chk",
        "response": [

            [
                "option",
                "Battery connect status",
                [
                    "Connected",
                    "Disconnected",
                ],
            ],
            ["int", "Battery capacity from BMS", "%"],
            [
                "option",
                "Battery force charging",
                [
                    "No",
                    "Yes",
                ],
            ],
            [
                "option",
                "Battery stop discharge flag",
                [
                    "Enable discharge",
                    "Disable discharge",
                ],
            ],
            [
                "option",
                "Battery stop charge flag",
                [
                    "Enable charge",
                    "Disable charge",
                ],
            ],
            ["float:r/10", "Battery bulk charging voltage from BMS", "V"],
            ["float:r/10", "Battery float charging voltage from BMS", "V"],
            ["float:r/10", "Battery cut-off voltage from BMS", "V"],
            ["float", "Battery max charging current", "A"],
            ["float", "Battery max discharge current", "A"]],
         "test_responses": [
            b"(0 100 0 0 1 532 532 450 0000 0030\x5E\n",
        ],
    },
}

SETTER_COMMANDS = {
    "PBT": {
        "name": "PBT",
        "description": "Set Battery Type",
        "help": " -- examples: PBT00 (set battery as AGM), Setting battery type, 00 for AGM, 01 for Flooded battery, 02 for user define, 03 for Pylontech, 04 for Shinheung, 05 for Weco, 06 for Soltaro, 07 for BAK, 08 for Lib, 09 for Lic",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PBT(0[0-9])$",
    },
}

class pi30m044(pi30max):
    def __str__(self):
        return "PI30 protocol handler for Voltronic Axpert Max 7.2k (general_model_name 044, model_name MKS2-7200)"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI30M044"
        # Add pi30m044 specific commands to pi30max commands
        self.COMMANDS.update(QUERY_COMMANDS)
        # Add pi30m044 specific setter commands
        self.COMMANDS.update(SETTER_COMMANDS)
        log.info(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")



""" usbport.py """
import logging
import os
import time

from powermon.dto.portDTO import PortDTO
from powermon.commands.result import Result, ResultType
from powermon.ports.abstractport import AbstractPort
from powermon.protocols import get_protocol_definition
from powermon.commands.command import Command

log = logging.getLogger("USBPort")


class USBPort(AbstractPort):
    """ usb port object """
    @classmethod
    def from_config(cls, config=None):
        log.debug("building usb port. config:%s", config)
        path = config.get("path", "/dev/hidraw0")
        # get protocol handler, default to PI30 if not supplied
        protocol = get_protocol_definition(protocol=config.get("protocol", "PI30"))
        return cls(path=path, protocol=protocol)

    def __init__(self, path, protocol) -> None:
        super().__init__(protocol=protocol)
        self.path = path
        self.port = None

    def to_dto(self):
        dto = PortDTO(type="usb", path=self.path, protocol=self.get_protocol().to_dto())
        return dto

    def is_connected(self) -> bool:
        return self.port is not None

    def connect(self) -> bool:
        if(self.is_connected()):
            log.debug("USBPort already connected")
            return True
        
        log.debug(f"USBPort connecting. path:{self.path}, protocol: {self.get_protocol()}")
        try:
            self.port = os.open(self.path, os.O_RDWR | os.O_NONBLOCK)
            log.debug(f"USBPort port number ${self.port}")
        except Exception as e:
            log.warning(f"Error openning usb port: {e}")
            self.error_message = e
        return self.is_connected()

    def disconnect(self) -> None:
        log.debug("USBPort disconnecting: %i", self.port)
        if self.port is not None:
            os.close(self.port)
        self.port = None

    def send_and_receive(self, command: Command) -> Result:
        if not self.is_connected():
            log.warning("USBPort not connected")
            return Result(command_code=command.code, result_type=ResultType.ERROR, raw_response=b"USBPort not connected")
        response_line = bytes()

        # Send the command to the open usb connection
        full_command = command.get_full_command()
        cmd_len = len(full_command)
        log.debug("length of to_send: %i", cmd_len)
        # for command of len < 8 it ok just to send
        # otherwise need to pack to a multiple of 8 bytes and send 8 at a time
        if cmd_len <= 8:
            # Send all at once
            log.debug("sending full_command in on shot")
            time.sleep(0.05)
            os.write(self.port, full_command)
        else:
            log.debug("multiple chunk send")
            chunks = [full_command[i:i + 8] for i in range(0, cmd_len, 8)]
            for chunk in chunks:
                # pad chunk to 8 bytes
                if len(chunk) < 8:
                    padding = 8 - len(chunk)
                    chunk += b'\x00' * padding
                log.debug("sending chunk: %s", chunk)
                time.sleep(0.05)
                os.write(self.port, chunk)
        time.sleep(0.25)
        # Read from the usb connection
        # try to a max of 100 times
        for x in range(100):
            # attempt to deal with resource busy and other failures to read
            try:
                time.sleep(0.15)
                r = os.read(self.port, 256)
                response_line += r
            except Exception as e:
                log.debug("USB read error: %s", e)
            # Finished is \r is in byte_response
            if bytes([13]) in response_line:
                # remove anything after the \r
                response_line = response_line[: response_line.find(bytes([13])) + 1]
                break
        log.debug("usb response was: %s", response_line)
        # response = self.get_protocol().check_response_and_trim(response_line)
        result = command.build_result(raw_response=response_line, protocol=self.get_protocol())

        return result



# Description: Output format for mqtt as individual topics
import logging
from powermon.formats.abstractformat import AbstractFormat
from powermon.commands.result import Result

log = logging.getLogger("Topics")


class Topics(AbstractFormat):
    def __init__(self, formatConfig, topic):
        super().__init__(formatConfig)
        self.name = "topics"
        self.results_topic = topic

    def sendsMultipleMessages(self) -> bool:
        return True

    def format(self, result: Result):
        log.info("Using output formatter: %s" % self.name)

        _result = []
        
        if len(result.get_responses()) == 0:
            return _result

        display_data = self.format_and_filter_data(result)
        log.debug(f"displayData: {display_data}")

        # Build array of mqtt messages
        msgs = []
        # Loop through responses build topics and messages
        for response in result.get_responses():
            value = response.get_data_value()
            unit = response.get_data_unit()
            name = response.get_data_name()
            log.debug(f"build_msgs: prefix {self.results_topic}, key {key}, value {value}, unit {unit}")
            msg = {"topic": f"{self.results_topic}/{name}/value", "payload": value}
            msgs.append(msg)
            if unit:
                msg = {"topic": f"{self.results_topic}/{name}/unit", "payload": unit}
                msgs.append(msg)
        log.debug(f"build_msgs: {msgs}")
        return msgs



import logging
import re

from .mqtt import mqtt
from ..helpers import get_kwargs
from ..helpers import key_wanted

log = logging.getLogger("domoticz_autodiscover")


class domoticz_autodiscover(mqtt):
    def __str__(self):
        return """outputs the to the supplied mqtt broker in hass format: eg "homeassistant/sensor/mpp_{tag}_{key}/state" """

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"__init__: kwargs {kwargs}")

    def build_msgs(self, *args, **kwargs):
        data = get_kwargs(kwargs, "data")
        tag = get_kwargs(kwargs, "tag")
        keep_case = get_kwargs(kwargs, "keep_case")

        filter = get_kwargs(kwargs, "filter")
        if filter is not None:
            filter = re.compile(filter)
        excl_filter = get_kwargs(kwargs, "excl_filter")
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)

        # Build array of mqtt messages with hass update format
        # assumes hass_config has been run
        # or hass updated manually
        msgs = []
        # Remove command and _command_description
        data.pop("_command", None)
        data.pop("_command_description", None)
        data.pop("raw_response", None)

        # Loop through responses
        for _key in data:
            # value = data[_key][0]
            unit = data[_key][1]
            # remove spaces
            key = _key.replace(" ", "_")
            if not keep_case:
                # make lowercase
                key = key.lower()
            if key_wanted(key, filter, excl_filter):
                #
                # CONFIG / AUTODISCOVER
                #
                # <discovery_prefix>/<component>/[<node_id>/]<object_id>/config
                # topic "homeassistant/binary_sensor/garden/config"
                # msg '{"name": "garden", "device_class": "motion", "state_topic": "homeassistant/binary_sensor/garden/state", "unit_of_measurement": "°C"}'
                topic = f"homeassistant/sensor/mpp_{tag}_{key}/config"
                topic = topic.replace(" ", "_")
                state_topic = f"domoticz/sensor/mpp_{tag}_{key}/state"
                state_topic = state_topic.replace(" ", "_")

                name = f"{tag} {_key}"
                if unit == "W":
                    payload = f'{{"name": "{name}", "stat_t": "{state_topic}", "unit_of_meas": "{unit}", "uniq_id": "mpp_{tag}_{key}", "stat_cla": "measurement", "device_class": "power"  }}'
                else:
                    payload = f'{{"name": "{name}", "stat_t": "{state_topic}", "unit_of_meas": "{unit}", "uniq_id": "mpp_{tag}_{key}"  }}'
                # msg = {"topic": topic, "payload": payload, "retain": True}
                msg = {"topic": topic, "payload": payload}
                msgs.append(msg)
                #
                # VALUE SETTING
                #
                # payload = value
                # msg = {"topic": state_topic, "payload": payload}
                # msgs.append(msg)
        return msgs



from argparse import ArgumentParser
import logging
import uvicorn
from fastapi import FastAPI, Request
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
# from fastapi_mqtt import FastMQTT, MQTTConfig

from mppsolar.version import __version__  # noqa: F401
from powermon import read_yaml_file

# Set-up logger
log = logging.getLogger("")
log.setLevel(logging.INFO)

# Initial some variables
templates = Jinja2Templates(directory="powermon_api/templates")
config = None
mqtt = None

# Create FastAPI app instance
app = FastAPI()
app.mount("/static", StaticFiles(directory="powermon_api/static"), name="static")


@app.get("/")
def read_root(request: Request):
    global config
    log.info(f"{config=}")
    return templates.TemplateResponse("home.html.j2", {"request": request, "config": config})


def main():
    """main entry point for powermon api"""
    global config, mqtt
    description = f"Power Device Monitoring Api, version: {__version__}"
    parser = ArgumentParser(description=description)

    parser.add_argument(
        "-C",
        "--configFile",
        type=str,
        help="Full location of yaml config file",
        default="./powermon-api.yaml",
    )
    args = parser.parse_args()

    # Build configuration from config file etc
    log.info("Using config file: %s", args.configFile)
    config = {"configFile": args.configFile}
    # build config with details from config file
    config.update(read_yaml_file(args.configFile))

    log.info(f"{config=}")
    # if "mqttbroker" in config:
    #     mqtt_config = MQTTConfig(
    #         host=config["mqttbroker"].get("name"),
    #         port=config["mqttbroker"].get("port"),
    #         keepalive=60,
    #         username=config["mqttbroker"].get("username"),
    #         password=config["mqttbroker"].get("password"),
    #     )
    #     try:
    #         mqtt = FastMQTT(config=mqtt_config)
    #         mqtt.init_app(app)
    #     except Exception:
    #         print("Unable to connect to mqttbroker")
    #         mqtt = None

    uvicornConfig = uvicorn.Config(
        "powermon_api.api:app",
        host=config["api"].get("host", "0.0.0.0"),
        port=config["api"].get("port", 5000),
        log_level=config["api"].get("log_level", "info"),
        reload=True,
    )
    server = uvicorn.Server(uvicornConfig)
    server.run()




import asyncio

from powermon.dto.deviceDTO import DeviceDTO
from powermon.dto.resultDTO import ResultDTO

class StateHandler(object):
    _instance = None
    _devices : dict[str,DeviceDTO] = {}
    _results : list[ResultDTO] = []
    _topics_waiting_for_result : dict[str, ResultDTO | None] = {}
    
    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = object.__new__(class_, *args, **kwargs)
        return class_._instance
    

    
    def recieved_result(self, topic: str, payload: bytes):
        
        print(f"Recieved result for {topic}")
        if self.is_command_result_topic(topic):
            result = ResultDTO.parse_raw(payload.decode())
            
            self._results.append(result)
            if topic in self._topics_waiting_for_result:
                self._topics_waiting_for_result[topic] = result
        
    async def get_next_result(self, topic: str) -> ResultDTO | None:
        print(f"Waiting for result for {topic}")
        print(self._topics_waiting_for_result)
        self._topics_waiting_for_result[topic] = None
        while True:
            if topic in self._topics_waiting_for_result:
                if self._topics_waiting_for_result[topic] is not None:
                    result = self._topics_waiting_for_result[topic]
                    del self._topics_waiting_for_result[topic]
                    return result
            else:
                raise RuntimeError("Topic not found in waiting for result list")
            print(f"Still waiting for result on {topic}")
            await asyncio.sleep(0.5)

    def is_command_result_topic(self, topic: str) -> bool:
        if "result" in topic:
            return True
        return False 

    def recieved_announcement(self, message) -> DeviceDTO:
        device = DeviceDTO.parse_raw(message)
        deviceId : str = device.device_id 
        self._devices[deviceId] = device
        return device


    def get_device_instances(self) -> list[DeviceDTO]:
        return list(self._devices.values())

    def get_device_instance(self, device_id) -> DeviceDTO | None:
        device = self._devices.get(device_id)
        return device

    async def get_results(self):
        return self._results


class CommandRequest:
    def __init__(self, command):
        self.command = command
        self.result = None
        self.done = False


from pydantic import BaseModel, Field


class ReadingDTO(BaseModel):
    data_name: str
    data_value: str
    data_unit: str = Field(default="")
    extra_info: dict = Field(default={})
    icon: str | None = Field(default=None)
    device_class: str | None = Field(default=None)
    state_class: str | None = Field(default=None)


from bluepy import btle
import logging


log = logging.getLogger("jkBleDelegate")


class jkBleDelegate(btle.DefaultDelegate):
    """
    BLE delegate to deal with notifications (information) from the JKBMS device
    """

    def __init__(self, jkbleio, protocol, record_type=None):
        btle.DefaultDelegate.__init__(self)
        # extra initialisation here
        self._jkbleio = jkbleio
        if protocol is None:
            print("ERROR")
            exit(1)
        self._protocol = protocol
        self._record_type = record_type
        self.notificationData = bytearray()

    def handleNotification(self, handle, data):
        # handle is the handle of the characteristic / descriptor that posted the notification
        # data is the data in this notification - may take multiple notifications to get all of a message
        log.debug("From handle: {:#04x} Got {} bytes of data".format(handle, len(data)))
        self.notificationData += bytearray(data)
        log.debug(f"Pre wipe to start {self.notificationData}")
        self.notificationData = self._protocol.wipe_to_start(self.notificationData)
        log.debug(f"Post wipe to start {self.notificationData}")
        # if not self._protocol.is_record_start(self.notificationData):
        #     log.debug(f"Not valid start of record - wiping data {self.notificationData}")
        #     self.notificationData = bytearray()
        if not self._protocol.is_record_correct_type(
            self.notificationData, self._record_type
        ):
            log.debug(
                f"Not expected type of record - wiping data {self.notificationData}"
            )
            self.notificationData = bytearray()
        if self._protocol.is_record_complete(self.notificationData):
            self._jkbleio.record = self.notificationData
            log.debug("record complete")
            self.notificationData = bytearray()



""" apicoordinator.py """
import logging
from time import time

from powermon.commands.command import Command
from powermon.commands.trigger import Trigger
from powermon.device import Device
from powermon.dto.commandDTO import CommandDTO
from powermon.dto.apicoordinatorDTO import ApicoordinatorDTO
from powermon.formats.simple import SimpleFormat
from powermon.outputs.api_mqtt import ApiMqtt

log = logging.getLogger("APICoordinator")


class ApiCoordinator:
    """ apicoordinator coordinates the api / mqtt interface """
    def __str__(self):
        if not self.enabled:
            return "ApiCoordinator DISABLED"
        return f"ApiCoordinator: adhocTopic: {self.adhoc_topic_format}, announceTopic: {self.announce_topic}"

    def to_dto(self):
        """ convert object to data transfer object """
        dto = ApicoordinatorDTO(name="ApiCoordinator", description="api coordinator to_dto is todo")
        return dto

    @classmethod
    def from_config(cls, config=None):
        """ build apicoordinator object from config dict """
        log.debug("ApiCoordinator config: %s", config)
        if not config:
            log.info("No api definition in config")
            refresh_interval = 300
            enabled = False
            announce_topic = "powermon/announce"
            adhoc_topic_format = "powermon/{device_id}/addcommand"
        else:
            refresh_interval = config.get("refresh_interval", 300)
            enabled = config.get("enabled", True)  # default to enabled if not specified
            announce_topic = config.get("announce_topic", "powermon/announce")
            adhoc_topic_format = config.get("adhoc_topic_format", "powermon/{device_id}/addcommand")

        return cls(adhoc_topic_format=adhoc_topic_format, announce_topic=announce_topic, enabled=enabled, refresh_interval=refresh_interval)

    def __init__(self, adhoc_topic_format: str, announce_topic: str, enabled: bool, refresh_interval: int):
        self.device = None
        self.mqtt_broker = None
        self.last_run = None
        self.adhoc_topic_format = adhoc_topic_format
        self.announce_topic = announce_topic
        self.refresh_interval = refresh_interval
        self.enabled = enabled

    def set_device(self, device: Device):
        """ store the device in the apicoordinator """
        self.device = device
        # self.announce(self.device)

    def set_mqtt_broker(self, mqtt_broker):
        """ store the mqtt broker in the apicoordinator """
        log.debug("setting mqtt_broker to %s", mqtt_broker)
        self.mqtt_broker = mqtt_broker

        if self.mqtt_broker is None or self.mqtt_broker.disabled:
            # no use having api running if no mqtt broker
            log.debug(self.mqtt_broker)
            log.debug("No mqttbroker (or it is disabled) so disabling ApiCoordinator")
            self.enabled = False
            return

        mqtt_broker.subscribe(self.get_addcommand_topic(), self.addcommand_callback)
        # mqtt_broker.publish(self.announceTopic, self.schedule.getScheduleConfigAsJSON())

    def get_addcommand_topic(self):
        """ get the adhoc topic """
        return self.adhoc_topic_format.format(device_id=self.device.device_id)

    def addcommand_callback(self, client, userdata, msg):
        """ add a callback """
        log.info("Client: %s, with userdata: %s, received `%s` on topic `%s`", client, userdata,msg.payload, msg.topic)
        json_string = msg.payload.decode("utf-8")
        log.debug("Yaml string: %s ", json_string)

        dto = CommandDTO.parse_raw(json_string)

        trigger = Trigger.from_DTO(dto.trigger)
        command = Command.from_DTO(dto)
        Command(code=dto.command_code, commandtype="basic", outputs=[], trigger=trigger)
        outputs = []

        output = ApiMqtt()
        output.set_formatter(SimpleFormat({}))
        outputs.append(output)

        command.set_outputs(outputs=outputs)
        command.set_mqtt_broker(self.mqtt_broker)

        self.device.add_command(command)

        return command

    def run(self):
        """ regular processing function, ensures that the announce isnt too frequent """
        if not self.enabled:
            return
        if not self.last_run or time() - self.last_run > self.refresh_interval:
            log.info("APICoordinator running")
            self.announce(self.device)
            self.last_run = time()

    def initialize(self):
        """ initialize the apicoordinator """
        if not self.enabled:
            return
        self.announce(self)

    def announce(self, obj):
        """ Announce jsonised obj dto to api """
        obj_dto = obj.to_dto()
        if not self.enabled:
            log.debug("Not announcing obj: %s as api DISABLED", obj_dto)
            return
        log.debug("Announcing obj: %s to api on topic: %s", obj_dto, self.announce_topic)
        self.mqtt_broker.publish(self.announce_topic, obj_dto.json())



import json as js
import logging
import re

from . import to_json
from .mqtt import mqtt
from ..helpers import get_kwargs

# from ..helpers import key_wanted

log = logging.getLogger("json_mqtt")


class json_mqtt(mqtt):
    def __str__(self):
        return "outputs all the results to the supplied mqtt broker in a single message formatted as json: eg "

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"__init__: kwargs {kwargs}")

    def build_msgs(self, *args, **kwargs):
        data = get_kwargs(kwargs, "data")
        tag = get_kwargs(kwargs, "tag")
        keep_case = get_kwargs(kwargs, "keep_case")
        mqtt_broker = get_kwargs(kwargs, "mqtt_broker")
        if mqtt_broker is not None:
            topic = mqtt_broker.results_topic
        else:
            topic = get_kwargs(kwargs, "mqtt_topic", default="mpp-solar")
        filter = get_kwargs(kwargs, "filter")
        if filter is not None:
            filter = re.compile(filter)
        excl_filter = get_kwargs(kwargs, "excl_filter")
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)

        # Build array of Influx Line Protocol II messages
        # Message format is: mpp-solar,command=QPGS0 max_charger_range=120.0
        #                    mpp-solar,command=inverter2 parallel_instance_number="valid"
        #                    measurement,tag_set field_set
        msgs = []
        # Remove command and _command_description
        cmd = data.pop("_command", None)
        data.pop("_command_description", None)
        data.pop("raw_response", None)
        if tag is None:
            tag = cmd
        output = to_json(data, keep_case, excl_filter, filter)
        payload = js.dumps(output)
        msg = {
             "topic": f"{tag}/{topic}",
            "payload": payload,
        }
        msgs.append(msg)
        return msgs



"""mqttbroker.py"""
import logging
from time import sleep

import paho.mqtt.client as mqtt_client

# Set-up logger
log = logging.getLogger("mqttbroker")


class MqttBroker:
    """
    Wrapper for mqtt broker connectivity and message proccessing
    """
    def __str__(self):
        if self.disabled:
            return "MqttBroker DISABLED"
        else:
            return f"MqttBroker name: {self.name}, port: {self.port}, user: {self.username}"

    @classmethod
    def from_config(cls, config={}):
        log.debug("mqttbroker config: %s", config)

        if config:
            name = config.get("name")
            port = config.get("port", 1883)
            username = config.get("username")
            password = config.get("password")
        else:
            name = None
            port = None
            username = None
            password = None

        return cls(name=name, port=port, username=username, password=password)

    def __init__(self, name, port=None, username=None, password=None):

        self.name = name
        self.port = port
        self.username = username
        self.password = password
        self._isConnected = False

        if self.name is None:
            self.disabled = True
        else:
            self.disabled = False
            self.mqttc = mqtt_client.Client()

    def on_connect(self, client, userdata, flags, rc):
        # 0: Connection successful
        # 1: Connection refused - incorrect protocol version
        # 2: Connection refused - invalid client identifier
        # 3: Connection refused - server unavailable
        # 4: Connection refused - bad username or password
        # 5: Connection refused - not authorised
        # 6-255: Currently unused.
        connection_result = [
            "Connection successful",
            "Connection refused - incorrect protocol version",
            "Connection refused - invalid client identifier",
            "Connection refused - server unavailable",
            "Connection refused - bad username or password",
            "Connection refused - not authorised",
        ]
        log.debug(f"MqttBroker connection returned result: {rc} {connection_result[rc]}")
        if rc == 0:
            self._isConnected = True
            return
        self._isConnected = False

    def on_disconnect(self, client, userdata, rc):
        print(f"Disconnection returned result: {rc}")
        self._isConnected = False

    def connect(self):
        if self.disabled:
            log.info(f"MQTT broker not enabled, was a broker name defined? '{self.name}'")
            return
        if not self.name:
            log.info(f"MQTT could not connect as no broker name '{self.name}'")
            return
        self.mqttc.on_connect = self.on_connect
        self.mqttc.on_disconnect = self.on_disconnect
        # if name is screen just return without connecting
        if self.name == "screen":
            # allows checking of message formats
            return
        try:
            log.debug(f"Connecting to {self.name} on port {self.port}")
            if self.username:
                # auth = {"username": mqtt_user, "password": mqtt_pass}
                _password = "********" if self.password is not None else "None"
                log.info(f"Using mqtt authentication, username: {self.username}, password: {_password}")
                self.mqttc.username_pw_set(self.username, password=self.password)
            else:
                log.debug("No mqtt authentication used")
                # auth = None
            self.mqttc.connect(self.name, port=self.port, keepalive=60)
            self.mqttc.loop_start()
            sleep(1)
        except ConnectionRefusedError as exc:
            log.warning(f"{self.name} refused connection '{exc}'")

    def start(self):
        if self.disabled:
            return
        if self._isConnected:
            self.mqttc.loop_start()

    def stop(self):
        log.debug("Stopping mqttbroker connection")
        if self.disabled:
            return
        if self.name:
            self.mqttc.loop_stop()
            if self._isConnected:
                log.debug("Disconnecting from mqtt broker")
                self.mqttc.disconnect

    def set(self, variable, value):
        setattr(self, variable, value)

    def update(self, variable, value):
        # only override if value is not None
        if value is None:
            return
        setattr(self, variable, value)

    def subscribe(self, topic, callback):
        # subscribe to mqtt topic
        if not self.name:
            return
        if self.disabled:
            return
        # check if connected, connect if not
        if not self._isConnected:
            log.debug("Not connected, connecting")
            self.connect()
        # Register callback
        self.mqttc.on_message = callback
        if self._isConnected:
            # Subscribe to command topic
            log.debug(f"Subscribing to topic {topic}")
            self.mqttc.subscribe(topic, qos=0)
        else:
            log.warning(f"Did not subscribe to topic {topic} as not connected to broker")

    def publishMultiple(self, data):
        if self.disabled:
            log.debug("Cannot publish msgs as mqttbroker disabled")
            return
        for msg in data:
            self.publish(msg["topic"], msg["payload"])

    def publish(self, topic, payload):
        if self.disabled:
            log.debug("Cannot publish msg as mqttbroker disabled")
            return
        #log.debug("Publishing '%s' to '%s'", payload, topic)
        if self.name == "screen":
            print(f"mqtt debug output only as broker name is 'screen' - topic: '{topic}', payload: '{payload}'")
            return
        # check if connected, connect if not
        if not self._isConnected:
            log.debug("Not connected, connecting")
            self.connect()
            sleep(1)
            if not self._isConnected:
                log.warning("mqtt broker did not connect")
                return
        try:
            infot = self.mqttc.publish(topic, payload)
            infot.wait_for_publish(5)
        except Exception as e:
            log.warning(str(e))

    def setAdhocCommands(self, config={}, callback=None):
        if not config:
            return
        if self.disabled:
            log.debug("Cannot setAdhocCommands as mqttbroker disabled")
            return

        adhoc_commands = config.get("adhoc_commands")
        # sub to command topic if defined
        adhoc_commands_topic = adhoc_commands.get("topic")
        if adhoc_commands_topic is not None:
            log.info(f"Setting adhoc commands topic to {adhoc_commands_topic}")
            self.subscribe(adhoc_commands_topic, callback)



""" errors.py - collection of powermon specific exceptions """


class ConfigError(Exception):
    """Exception for invaild configurations"""


class PowermonProtocolError(Exception):
    """Exception for errors with protocol definitions"""


class PowermonWIP(Exception):
    """Exception for work not yet done"""



""" commands / command.py """
import logging

from powermon.commands.command_definition import CommandDefinition
# from powermon.commands.result import ResultType
from powermon.commands.parameter import Parameter
# from powermon.commands.reading import Reading
from powermon.commands.reading_definition import ReadingDefinition
from powermon.commands.result import Result
from powermon.commands.trigger import Trigger
from powermon.dto.commandDTO import CommandDTO
from powermon.outputs import OutputType, getOutputs
from powermon.outputs.abstractoutput import AbstractOutput
from powermon.outputs.api_mqtt import ApiMqtt

# from time import localtime, strftime


log = logging.getLogger("Command")


class Command():
    """
    Command object, holds the details of the command, including:
    - trigger
    - outputs
    """
    def __str__(self):
        if self.code is None:
            return "empty command object"

        last_run = self.trigger.get_last_run()
        next_run = self.trigger.get_next_run()

        _outs = ""
        for output in self.outputs:
            _outs += str(output)

        return f"Command: {self.code=} {self.full_command=}, {self.type=}, \
            [{_outs=}], {last_run=}, {next_run=}, {str(self.trigger)}, {self.command_definition=}"

    def __init__(self, code: str, commandtype: str, outputs: list[AbstractOutput], trigger: Trigger):
        self.code = code
        self.command_description = "Not set"
        self.type = commandtype
        self.set_outputs(outputs)

        self.trigger: Trigger = trigger

        self.full_command = None
        self.command_definition: CommandDefinition = None
        self.device_id = None  # TODO: shouldnt need this
        log.debug(self)

    @classmethod
    def from_config(cls, config=None) -> "Command":
        """build object from config dict"""
        # need to have a config defined
        # minimum is
        # - command: QPI
        if not config:
            log.warning("Invalid command config")
            raise TypeError("Invalid command config")
            # return None

        code = config.get("command")
        if code is None:
            log.info("command must be defined")
            raise TypeError("command must be defined")
        commandtype = config.get("type", "basic")
        outputs = getOutputs(config.get("outputs", ""))
        trigger = Trigger.fromConfig(config=config.get("trigger"))
        return cls(code=code, commandtype=commandtype, outputs=outputs, trigger=trigger)

    def build_result(self, raw_response=None, protocol=None) -> Result:
        log.debug(f"build_result: code:{self.code}, command_definition:{self.command_definition}")
        trimmed_response = protocol.check_response_and_trim(raw_response)
        result = Result(
            self.code, result_type=self.command_definition.result_type,
            reading_definitions=self.get_reading_definitions(),
            parameters=self.command_definition.parameters, raw_response=trimmed_response
        )
        return result

    def set_full_command(self, full_command):
        """store the full command"""
        self.full_command = full_command

    def get_full_command(self) -> str | None:
        """return the full command, including CRC and/or headers"""
        return self.full_command

    def set_command_definition(self, command_definition: CommandDefinition):
        """store the definition of the command"""
        if command_definition is None:
            raise ValueError("CommandDefinition cannot be None")

        # Check if the definition is valid for the command
        if command_definition.is_command_code_valid(self.code) is False:
            raise ValueError(f"Command code {self.code} is not valid for command definition regex {command_definition.regex}")
        self.command_definition = command_definition
        self.command_description = command_definition.description

        # set command description in each of the outputs
        # QUESTION: why, cant we just pass the command object?
        for output in self.outputs:
            output.formatter.set_command_description(self.command_description)

    def get_reading_definitions(self) -> list[ReadingDefinition]:
        return self.command_definition.reading_definitions

    def set_outputs(self, outputs: list[AbstractOutput]):
        self.outputs = outputs
        for output in self.outputs:
            output.set_command(self.code)  # TODO: shouldnt need this

    def set_device_id(self, device_id):  # TODO: shouldnt need this
        self.device_id = device_id
        for output in self.outputs:
            output.set_device_id(device_id)

    def get_parameters(self) -> dict[str, Parameter]:
        return self.command_definition.parameters



    @classmethod
    def from_DTO(cls, command_dto: CommandDTO) -> "Command":
        trigger = Trigger.from_DTO(command_dto.trigger)
        command = cls(code=command_dto.command_code, commandtype="basic", outputs=[], trigger=trigger)
        outputs = []
        for output_dto in command_dto.outputs:
            if output_dto.type == OutputType.API_MQTT:
                outputs.append(ApiMqtt.from_DTO(output_dto))
        command.set_outputs(outputs=outputs)
        return command

    def is_due(self):
        """is this command due to run?"""
        return self.trigger.is_due()

    def touch(self):
        self.trigger.touch()

    def to_dto(self):
        return CommandDTO(
            command_code=self.code,
            device_id=self.device_id,
            result_topic=self.outputs[0].get_topic(),
            trigger=self.trigger.to_dto(),
            outputs=[output.to_dto() for output in self.outputs],
        )



import logging

from .baseoutput import baseoutput
from ..helpers import get_kwargs

log = logging.getLogger("raw")


class raw(baseoutput):
    def __str__(self):
        return "outputs the raw results to standard out"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"__init__: kwargs {kwargs}")

    def output(self, *args, **kwargs):
        log.info("Using output processor: raw")
        log.debug(f"kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")
        if data is None:
            return
        _desc = "No description found"
        if "_command_description" in data:
            _desc = data["_command_description"]
            del data["_command_description"]
        if "_command" in data:
            print(f"Command: {data['_command']} - {_desc}")
            print("-" * 60)
            del data["_command"]
        if "raw_response" in data:
            key = "raw_response"
            value = data[key][0]
            print(f"{key:<30}\t{value!a:<15}")
        return



import logging
import socket
import time

from .baseio import BaseIO
from ..helpers import get_kwargs

log = logging.getLogger("remoteSocketIO")


class remoteSocketIO(BaseIO):
    def __init__(self, *args, **kwargs) -> None:
        self._remote_ip = get_kwargs(kwargs, "remote_ip")
        self._remote_port = get_kwargs(kwargs, "remote_port")

    def send_and_receive(self, *args, **kwargs) -> dict:
        full_command = get_kwargs(kwargs, "full_command")
        response_line = None
        log.debug(f"host ip: {self._remote_ip}, host port: {self._remote_port}")

        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.connect((self._remote_ip, self._remote_port))
                log.debug("Executing command via remoteserialio...")
                s.send(full_command)
                time.sleep(0.1)  # give serial port time to receive the data
                response_line = s.recv(1024)
                # response_line = s.read_until(b"\r")
                log.debug("socket response was: %s", response_line)
                return response_line
        except Exception as e:
            log.warning(f"socket read error: {e}")
        log.info("Command execution failed")
        return {"ERROR": ["Socket command execution failed", ""]}



""" reading_definition.py """
import calendar  # needed for INFO type evaluating templates
import logging
from abc import ABC, abstractmethod
from enum import auto

from strenum import LowercaseStrEnum

from powermon.commands.reading import Reading

log = logging.getLogger("ReadingDefinition")


class ResponseType(LowercaseStrEnum):
    """
    the type of the response
    - determines how to read and translate to useful info
    """
    ACK = auto()
    INT = auto()
    OPTION = auto()
    BYTES = "bytes.decode"  # can't use auto() for this value
    FLOAT = auto()
    ENFLAGS = auto()
    STRING = auto()
    FLAGS = auto()
    INFO = auto()


class ReadingType(LowercaseStrEnum):
    """
    the type of the reading
    - higher level type, like Wh etc
    - allows translations
    """
    ACK = auto()
    WATT_HOURS = auto()
    WATTS = auto()
    TIME = auto()
    MESSAGE = auto()
    FLAG = auto()
    AMPERAGE = auto()
    TEMPERATURE = auto()
    PERCENTAGE = auto()
    FREQUENCY = auto()
    AMPS = auto()


class ReadingDefinition(ABC):
    """
    Create a flat representation to check if a response is valid for the command.
    It doesn't contain the response value, just the definition of what is valid.
    """
    def __str__(self):
        return f"{self.index=}, {self.name=}, {self.description=}, {self.response_type=}, {self.unit=}, instance={type(self)}"

    def __init__(self, index, name, response_type, description, device_class, state_class, icon, unit=""):
        self.index = index
        self.name = name
        self.response_type = response_type
        self.unit = unit
        self.description = description
        self.device_class = device_class
        self.state_class = state_class
        self.icon = icon

    def get_description(self) -> str:
        return self.description

    @abstractmethod
    def reading_from_raw_response(self, raw_value) -> list[Reading]:
        raise NotImplementedError

    def get_invalid_message(self, raw_value) -> str:
        return f"Invalid response for {self.get_description()}: {raw_value}"

    def is_info(self) -> bool:
        return False

    @classmethod
    def multiple_from_config(cls, reading_definitions_config: list[dict]) -> dict[int, "ReadingDefinition"]:
        if reading_definitions_config is None:
            return {}
        else:
            reading_definitions: dict[int, "ReadingDefinition"] = {}
            for i, reading_definition_config in enumerate(reading_definitions_config):
                reading_definition = cls.from_config(reading_definition_config, i)
                log.debug("reading definition: %s", reading_definition)
                reading_definitions[reading_definition.index] = reading_definition
            return reading_definitions

    @classmethod
    def from_config(cls, reading_definition_config: dict, i) -> "ReadingDefinition":
        index = i
        name = reading_definition_config.get("name")
        description = reading_definition_config.get("description")
        response_type = reading_definition_config.get("response_type")
        reading_type = reading_definition_config.get("reading_type")
        device_class = reading_definition_config.get("device-class", None)
        state_class = reading_definition_config.get("state-class", None)
        icon = reading_definition_config.get("icon", None)

        match reading_type:
            case ReadingType.ACK:
                return ReadingDefinitionACK(
                    index=index,
                    name=name,
                    response_type=response_type,
                    description=description,
                    device_class=device_class,
                    state_class=state_class,
                    icon=icon,
                )
            case ReadingType.WATT_HOURS:
                return ReadingDefinitionWattHours(
                    index=index,
                    name=name,
                    response_type=response_type,
                    description=description,
                    device_class=device_class,
                    state_class=state_class,
                    icon=icon,
                )
            case ReadingType.MESSAGE:
                options = None
                if response_type == ResponseType.OPTION:
                    options: dict[str, str] = reading_definition_config.get("options")
                return ReadingDefinitionMessage(
                    index=index,
                    name=name,
                    response_type=response_type,
                    description=description,
                    options=options,
                    device_class=device_class,
                    state_class=state_class,
                    icon=icon,
                )
            case ReadingType.TEMPERATURE:
                return ReadingDefinitionTemperature(
                    index=index,
                    name=name,
                    response_type=response_type,
                    description=description,
                    device_class=device_class,
                    state_class=state_class,
                    icon=icon
                )
            case ReadingType.TIME:
                return ReadingDefinitionDefault(
                    index=index,
                    name=name,
                    response_type=response_type,
                    description=description,
                    device_class=device_class,
                    state_class=state_class,
                    icon=icon,
                    unit="s"
                )
            case ReadingType.FLAG:
                return ReadingDefinitionDefault(
                    index=index,
                    name=name,
                    response_type=response_type,
                    description=description,
                    device_class=device_class,
                    state_class=state_class,
                    icon=icon
                )
            case ReadingType.AMPERAGE:
                return ReadingDefinitionDefault(
                    index=index,
                    name=name,
                    response_type=response_type,
                    description=description,
                    device_class=device_class,
                    state_class=state_class,
                    icon=icon,
                    unit="A"
                )
            case ReadingType.PERCENTAGE:
                return ReadingDefinitionDefault(
                    index=index,
                    name=name,
                    response_type=response_type,
                    description=description,
                    device_class=device_class,
                    state_class=state_class,
                    icon=icon,
                    unit="%"
                )
            case ReadingType.FREQUENCY:
                return ReadingDefinitionDefault(
                    index=index,
                    name=name,
                    response_type=response_type,
                    description=description,
                    device_class=device_class,
                    state_class=state_class,
                    icon=icon,
                    unit="Hz"
                )
            case ReadingType.WATTS:
                return ReadingDefinitionDefault(
                    index=index,
                    name=name,
                    response_type=response_type,
                    description=description,
                    device_class=device_class,
                    state_class=state_class,
                    icon=icon,
                    unit="W"
                )

            case _:
                raise ValueError(
                    f"Reading description: {description} has unknown reading_type definition type: {reading_type}"
                )


class ReadingDefinitionDefault(ReadingDefinition):
    def __init__(
        self,
        index: int,
        name: str,
        response_type: str,
        description: str,
        device_class: str = None,
        state_class: str = None,
        icon: str = None,
        unit: str = ""
    ):
        super().__init__(index, name, response_type, description, device_class, state_class, icon, unit=unit)

    def translate_raw_response(self, raw_value) -> str:
        return raw_value.decode()

    def reading_from_raw_response(self, raw_value) -> list[Reading]:
        value = self.translate_raw_response(raw_value)
        return [
            Reading(
                data_name=self.description,
                data_value=value,
                data_unit=self.unit,
                device_class=self.device_class,
                state_class=self.state_class,
                icon=self.icon,
            )
        ]


class ReadingDefinitionACK(ReadingDefinition):
    def __init__(
        self,
        index: int,
        name: str,
        response_type: str,
        description: str,
        device_class: str = None,
        state_class: str = None,
        icon: str = None,
    ):
        super().__init__(index, name, response_type, description, device_class, state_class, icon)

        self.fail_code = "NAK"
        self.fail_description = "Failed"
        self.success_code = "ACK"
        self.success_description = "Successful"

    def reading_from_raw_response(self, raw_value) -> list[Reading]:
        value = raw_value.decode()
        if value == self.success_code:
            return [
                Reading(
                    data_name=self.description,
                    data_value=self.success_description,
                    data_unit=None,
                    device_class=self.device_class,
                    state_class=self.state_class,
                    icon=self.icon,
                )
            ]
        elif value == self.fail_code:
            return [
                Reading(
                    data_name=self.description,
                    data_value=self.fail_description,
                    data_unit=None,
                    device_class=self.device_class,
                    state_class=self.state_class,
                    icon=self.icon,
                )
            ]

    def get_description(self) -> str:
        return self.description

class ReadingDefinitionWattHours(ReadingDefinition):
    def __init__(
        self,
        index: int,
        name: str,
        response_type: str,
        description: str,
        device_class: str = None,
        state_class: str = None,
        icon: str = None
    ):
        super().__init__(index, name, response_type, description, device_class, state_class, icon, unit="Wh")
        if response_type not in [ResponseType.INT]:
            raise TypeError(f"Wh response must be of type int, ResponseType {response_type} is not valid")

    def translate_raw_response(self, raw_value) -> str:
        return int(raw_value.decode())

    def reading_from_raw_response(self, raw_value) -> list[Reading]:
        value = self.translate_raw_response(raw_value)
        return [
            Reading(
                data_name=self.description,
                data_value=value,
                data_unit=self.unit,
                device_class=self.device_class,
                state_class=self.state_class,
                icon=self.icon,
            )
        ]


class ReadingDefinitionMessage(ReadingDefinition):
    def __init__(
        self,
        index: int,
        name: str,
        response_type: str,
        description: str,
        options: dict[str, str] = None,
        device_class: str = None,
        state_class: str = None,
        icon: str = None
    ):
        super().__init__(index, name, response_type, description, device_class, state_class, icon, unit="")
        if response_type == ResponseType.OPTION and not isinstance(options, dict):
            raise TypeError(f"For Reading Defininition {self.name}, options must be a dict if response_type is OPTION")
        
        self.options = options

    def translate_raw_response(self, raw_value) -> str:
        if self.response_type == ResponseType.OPTION:
            value = str(raw_value.decode())
            print(f"Reading:{self.description} Value:{value}")
            
            return self.options[value]
        return raw_value.decode('utf-8')

    def reading_from_raw_response(self, raw_value) -> list[Reading]:
        value = self.translate_raw_response(raw_value)
        return [
            Reading(
                data_name=self.description,
                data_value=str(value),
                data_unit=self.unit,
                device_class=self.device_class,
                state_class=self.state_class,
                icon=self.icon,
            )
        ]

class ReadingDefinitionTemperature(ReadingDefinition):
    def __init__(self, index: int, name: str, description: str,  response_type: ResponseType, device_class: str = None, state_class: str = None, icon: str = None):
        #TODO: find a way to make the unit configurable
        super().__init__(index, name, response_type, description, device_class, state_class, icon, unit="°C")
        if response_type not in [ResponseType.INT, ResponseType.FLOAT]:
            raise TypeError(f"Temperature response must be of type int or float, ResponseType {response_type} is not valid")

    def translate_raw_response(self, raw_value) -> float:
        return float(raw_value.decode())

    def reading_from_raw_response(self, raw_value) -> list[Reading]:
        value = self.translate_raw_response(raw_value)
        return [
            Reading(
                data_name=self.description,
                data_value=str(value),
                data_unit=self.unit,
                device_class=self.device_class,
                state_class=self.state_class,
                icon=self.icon,
            )
        ]

    def get_description(self) -> str:
        return self.description

class ReadingDefinitionENFlags(ReadingDefinition):
    def __init__(
        self, index: int, description: str, enflags: dict[str, dict[str, str]], device_class: str = None, state_class: str = None, icon: str = None
    ):
        self.index = index
        self.description = description
        self.enflags = enflags  # what does enflags mean?
        self.device_class = device_class
        self.state_class = state_class
        self.icon = icon

    def translate_raw_response(self, raw_value) -> dict[str, str]:
        return_values = {}
        status = "unknown"
        for i, item in enumerate(raw_value):
            item = chr(item)
            if item == "E":
                status = "enabled"
            elif item == "D":
                status = "disabled"
            else:
                if item in self.enflags:
                    _key = self.enflags[item]["name"]
                else:
                    _key = f"unknown_{i}"
                return_values[_key] = status
        return return_values

    def reading_from_raw_response(self, raw_value) -> list[Reading]:
        values: dict[str, str] = self.translate_raw_response(raw_value)
        responses = []
        for name, value in values.items():
            responses.append(
                Reading(
                    data_name=name, data_value=value, data_unit="", device_class=self.device_class, state_class=self.state_class, icon=self.icon
                )
            )

        return responses

    def get_description(self) -> str:
        return self.description

class ReadingDefinitionFlags(ReadingDefinition):
    def __init__(self, index: int, description: str, flags: list[str], device_class: str = None, state_class: str = None, icon: str = None):
        self.index = index
        self.description = description
        self.flags = flags
        self.device_class = device_class
        self.state_class = state_class
        self.icon = icon

    def translate_raw_response(self, raw_value) -> dict[str, int]:
        return_value = {}
        for i, flag in enumerate(raw_value):
            if self.flags[i]:  # only append value if flag name is present
                return_value[self.flags[i]] = int(chr(flag))
        return return_value

    def reading_from_raw_response(self, raw_value) -> list[Reading]:
        values: dict[str, int] = self.translate_raw_response(raw_value)
        responses = []
        for name, value in values.items():
            responses.append(
                Reading(
                    data_name=name, data_value=value, data_unit="bool", device_class=self.device_class, state_class=self.state_class, icon=self.icon
                )
            )
        return responses

    def get_description(self) -> str:
        return self.description





from typing import Optional

from pydantic import BaseModel

from .portDTO import PortDTO
from .commandDTO import CommandDTO


class DeviceDTO(BaseModel):
    device_id: str
    model: Optional[str]
    manufacturer: Optional[str]
    port: PortDTO
    commands: list[CommandDTO]



import logging
import serial
import time

from .baseio import BaseIO
from ..helpers import get_kwargs

log = logging.getLogger("DalySerialIO")


class DalySerialIO(BaseIO):
    def __init__(self, *args, **kwargs) -> None:
        self._serial_port = get_kwargs(kwargs, "device_path")
        self._serial_baud = get_kwargs(kwargs, "serial_baud")

    def send_and_receive(self, *args, **kwargs) -> dict:
        full_command = get_kwargs(kwargs, "full_command")
        response_line = b""
        log.debug(f"port {self._serial_port}, baudrate {self._serial_baud}")
        try:
            with serial.serial_for_url(self._serial_port, self._serial_baud) as s:
                log.debug("Executing command via dalyserialio...")
                s.timeout = 0.5
                s.write_timeout = 1
                s.reset_input_buffer()
                s.reset_output_buffer()
                s.write(full_command)
                # read until no more data
                while True:
                    time.sleep(0.5)  # give serial port time to receive the data
                    to_read = s.in_waiting
                    log.debug(f"bytes waiting {to_read}")
                    if to_read == 0:
                        break
                    # got some data to read
                    response_line += s.read(to_read)

                log.debug("serial response was: %s", response_line)
                return response_line
        except Exception as e:
            log.warning(f"Serial read error: {e}")
        log.info("Command execution failed")
        return {"ERROR": ["Serial command execution failed", ""]}



import logging

from .pi30max import pi30max

log = logging.getLogger("pi30m045")

QUERY_COMMANDS = {
    "QDI": {
        "name": "QDI",
        "description": "Default Settings inquiry",
        "type": "QUERY",
        "response": [
            ["float", "AC Output Voltage", "V"],
            ["float", "AC Output Frequency", "Hz"],
            ["int", "Max AC Charging Current", "A"],
            ["float", "Battery Under Voltage", "V"],
            ["float", "Battery Float Charge Voltage", "V"],
            ["float", "Battery Bulk Charge Voltage", "V"],
            ["float", "Battery Recharge Voltage", "V"],
            ["int", "Max Charging Current", "A"],
            ["option", "Input Voltage Range", ["Appliance", "UPS"]],
            [
                "option",
                "Output Source Priority",
                ["Utility first", "Solar first", "SBU first"],
            ],
            [
                "option",
                "Charger Source Priority",
                [
                    "Solar first",
                    "Solar + Utility",
                    "Only solar charging permitted",
                ],
            ],
            [
                "option",
                "Battery Type",
                [
                    "AGM",
                    "Flooded",
                    "User",
                    "TBD",
                    "Pylontech",
                    "WECO",
                    "Soltaro",
                    "LIb-protocol compatible",
                    "3rd party Lithium",
                ],
            ],
            ["option", "Buzzer", ["enabled", "disabled"]],
            ["option", "Power saving", ["disabled", "enabled"]],
            ["option", "Overload restart", ["disabled", "enabled"]],
            ["option", "Over temperature restart", ["disabled", "enabled"]],
            ["option", "LCD Backlight", ["disabled", "enabled"]],
            ["option", "Primary source interrupt alarm", ["disabled", "enabled"]],
            ["option", "Record fault code", ["disabled", "enabled"]],
            ["option", "Overload bypass", ["disabled", "enabled"]],
            ["option", "LCD reset to default", ["disabled", "enabled"]],
            [
                "option",
                "Output mode",
                [
                    "single machine",
                    "parallel output",
                    "Phase 1 of 3 phase output",
                    "Phase 2 of 3 phase output",
                    "Phase 3 of 3 phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output (120°)",
                    "Phase 2 of 2 phase output (180°)",
                    "Unknown Output Mode",
                ],
            ],
            ["float", "Battery Redischarge Voltage", "V"],
            [
                "option",
                "PV OK condition",
                [
                    "As long as one unit of inverters has connect PV, parallel system will consider PV OK",
                    "Only All of inverters have connect PV, parallel system will consider PV OK",
                ],
            ],
            [
                "option",
                "PV Power Balance",
                [
                    "PV input max current will be the max charged current",
                    "PV input max power will be the sum of the max charged power and loads power",
                ],
            ],
            ["int", "Max Charging Time at CV", "min"],
            ["int", "Max Discharging current", "A"],
        ],
        "test_responses": [
            b"(230.0 50.0 0030 44.0 54.0 56.4 46.0 60 0 0 2 0 0 0 0 0 1 1 1 0 1 0 54.0 0 1 224\xeb\xbc\r",
        ],
    },
    "QPGS": {
        "name": "QPGS",
        "description": "Parallel Information inquiry",
        "help": " -- example: QPGS0 queries the values of various metrics from instance 0 of parallel setup Inverters",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [1, "Parallel instance number", "option", ["Not valid", "valid"]],
            [2, "Serial number", "bytes:r.decode()", "", {"icon": "mdi:identifier"}],
            [
                3,
                "Work mode",
                "str_keyed",
                {
                    "P": "Power On Mode",
                    "S": "Standby Mode",
                    "L": "Line Mode",
                    "B": "Battery Mode",
                    "F": "Fault Mode",
                    "H": "Power Saving Mode",
                    "D": "Shutdown Mode",
                },
            ],
            [
                4,
                "Fault code",
                "str_keyed",
                {
                    "00": "No fault",
                    "01": "Fan is locked",
                    "02": "Over temperature",
                    "03": "Battery voltage is too high",
                    "04": "Battery voltage is too low",
                    "05": "Output short circuited or Over temperature",
                    "06": "Output voltage is too high",
                    "07": "Over load time out",
                    "08": "Bus voltage is too high",
                    "09": "Bus soft start failed",
                    "10": "PV over current",
                    "11": "PV over voltage",
                    "12": "DC over current",
                    "13": "Battery discharge over current",
                    "51": "Over current inverter",
                    "52": "Bus voltage too low",
                    "53": "Inverter soft start failed",
                    "54": "Self-test failed",
                    "55": "Over DC voltage on output of inverter",
                    "56": "Battery connection is open",
                    "57": "Current sensor failed",
                    "58": "Output voltage is too low",
                    "60": "Power feedback protection",
                    "71": "Firmware version different",
                    "72": "Current sharing fault",
                    "80": "CAN communication failed",
                    "81": "Parallel host line lost",
                    "82": "Parallel synchronized signal lost",
                    "83": "Parallel battery voltage detect different",
                    "84": "AC input voltage or frequency detected different",
                    "85": "AC output current unbalanced",
                    "86": "AC output mode setting different",
                },
            ],
            [5, "Grid Voltage", "float", "V", {"icon": "mdi:power-plug", "device-class": "voltage"}],
            [6, "Grid Frequency", "float", "Hz", {"icon": "mdi:current-ac", "device-class": "frequency"}],
            [7, "AC Output Voltage", "float", "V", {"icon": "mdi:power-plug", "device-class": "voltage"}],
            [8, "AC Output Frequency", "float", "Hz", {"icon": "mdi:current-ac", "device-class": "frequency"}],
            [9, "AC Output Apparent Power", "int", "VA", {"icon": "mdi:power-plug", "device-class": "apparent_power"}],
            [10, "AC Output Active Power", "int", "W", {"icon": "mdi:power-plug", "device-class": "power", "state_class": "measurement"}],
            [11, "Load Percentage", "int", "%", {"icon": "mdi:brightness-percent"}],
            [12, "Battery Voltage", "float", "V", {"icon": "mdi:battery-outline", "device-class": "voltage"}],
            [13, "Battery Charging Current", "int", "A", {"icon": "mdi:current-dc", "device-class": "current"}],
            [14, "Battery Capacity", "int", "%", {"device-class": "battery"}],
            [15, "PV1 Input Voltage", "float", "V", {"icon": "mdi:solar-power", "device-class": "voltage"}],
            [
                16,
                "Total Charging Current",
                "int",
                "A",
                {"icon": "mdi:brightness-percent", "device-class": "current"},
            ],
            [
                17,
                "Total AC Output Apparent Power",
                "int",
                "VA",
                {"icon": "mdi:power-plug", "device-class": "apparent_power"},
            ],
            [
                18,
                "Total Output Active Power",
                "int",
                "W",
                {"icon": "mdi:power-plug", "device-class": "power", "state_class": "measurement"},
            ],
            [
                19,
                "Total AC Output Percentage",
                "int",
                "%",
                {"icon": "mdi:brightness-percent"},
            ],
            [
                20,
                "Inverter Status",
                "flags",
                [
                    "Is SCC OK",
                    "Is AC Charging",
                    "Is SCC Charging",
                    "Is Battery Over Voltage",
                    "Is Battery Under Voltage",
                    "Is Line Lost",
                    "Is Load On",
                    "Is Configuration Changed",
                ],
            ],
            [
                21,
                "Output mode",
                "option",
                [
                    "single machine",
                    "parallel output",
                    "Phase 1 of 3 phase output",
                    "Phase 2 of 3 phase output",
                    "Phase 3 of 3 phase output",
                    "Phase 1 of 2 phase output",
                    "Phase 2 of 2 phase output (120°)",
                    "Phase 2 of 2 phase output (180°)",
                    "Unknown Output Mode",
                ],
            ],
            [
                22,
                "Charger source priority",
                "option",
                ["Solar first", "Solar + Utility", "Solar only"],
            ],
            [23, "Max Charger Current", "int", "A", {"device-class": "current"}],
            [24, "Max Charger Range", "int", "A", {"device-class": "current"}],
            [25, "Max AC Charger Current", "int", "A", {"device-class": "current"}],
            [26, "PV1 Input Current", "int", "A", {"icon": "mdi:solar-power", "device-class": "power"}],
            [
                27,
                "Battery Discharge Current",
                "int",
                "A",
                {"icon": "mdi:battery-negative", "device-class": "current"},
            ],
            [28, "PV2 Input Voltage", "float", "V", {"icon": "mdi:solar-power", "device-class": "voltage"}],
            [29, "PV2 Input Current", "int", "A", {"icon": "mdi:solar-power", "device-class": "current"}],
        ],
        "test_responses": [
            b"(0 92932105105315 B 00 000.0 00.00 230.0 50.00 0989 0907 012 53.2 009 090 349.8 009 00989 00907 011 10100110 0 1 100 120 030 02 000 275.3 02i]\r",
        ],
        "regex": "QPGS(\\d+)$",
    },
}
SETTER_COMMANDS = {
    "DAT": {
        "name": "DAT",
        "description": "Set Date Time",
        "help": " -- examples: DATYYMMDDHHMMSS (12 digits after DAT)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "DAT(\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d\\d)$",
    },
    "PCP": {
        "name": "PCP",
        "description": "Set Device Charger Priority",
        "help": " -- examples: PCP00 (set solar first), PCP01 (set solar and utility), PCP02 (set solar only charging)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PCP(0[0123])$",
    },
    "PPCP": {
        "name": "PPCP",
        "description": "Set Parallel Device Charger Priority",
        "help": " -- examples: PPCP000 (set unit 1 to 00 - solar first), PPCP101 (set unit 1 to 01 - solar and utility), PPCP202 (set unit 2 to 02 - solar only charging)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "PPCP(\\d0[0123])$",
    },
    "MUCHGC": {
        "name": "MUCHGC",
        "description": "Set Utility Max Charging Current",
        "help": " -- example: MUCHGCuaaa (MUCHGC1030 set unit 1 utility max charging current to 30A)",
        "type": "SETTER",
        "response": [["ack", "Command execution", {"NAK": "Failed", "ACK": "Successful"}]],
        "test_responses": [
            b"(NAK\x73\x73\r",
            b"(ACK\x39\x20\r",
        ],
        "regex": "MUCHGC(\\d\\d\\d\\d)$",
    },
}

class pi30m045(pi30max):
    def __str__(self):
        return "PI30 protocol handler for LV6548 and similar inverters"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI30M045"
        # Add pi30m045 specific commands to pi30max commands
        self.COMMANDS.update(QUERY_COMMANDS)
        # Add pi30m045 specific setter commands
        self.COMMANDS.update(SETTER_COMMANDS)
        log.info(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")



import logging
import re

from .baseoutput import baseoutput
from ..helpers import get_kwargs, key_wanted

log = logging.getLogger("simpleunits")


class simpleunits(baseoutput):
    def __str__(self):
        return "outputs 'param=value(unit)' only to standard out"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"processor.value __init__ kwargs {kwargs}")

    def printHeader(command, description):
        pass

    def output(self, *args, **kwargs):
        log.info("Using output processor: value")
        log.debug(f"kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")
        if data is None:
            return

        # check if config supplied
        config = get_kwargs(kwargs, "config")
        if config is not None:
            log.debug(f"config: {config}")
            # get formatting info
            remove_spaces = config.get("remove_spaces", True)
            keep_case = config.get("keep_case", False)
            filter = config.get("filter", None)
            excl_filter = config.get("excl_filter", None)
        else:
            # get formatting info
            remove_spaces = True
            keep_case = get_kwargs(kwargs, "keep_case")
            filter = get_kwargs(kwargs, "filter")
            excl_filter = get_kwargs(kwargs, "excl_filter")

        if filter is not None:
            filter = re.compile(filter)
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)

        # remove raw response
        if "raw_response" in data:
            data.pop("raw_response")

        # build header
        if "_command" in data:
            data.pop("_command")

        if "_command_description" in data:
            data.pop("_command_description")

        # build data to display
        displayData = {}
        for key in data:
            _values = data[key]
            # remove spaces
            if remove_spaces:
                key = key.replace(" ", "_")
            if not keep_case:
                # make lowercase
                key = key.lower()
            if key_wanted(key, filter, excl_filter):
                displayData[key] = _values
        log.debug(f"displayData: {displayData}")

        # print data
        for key in displayData:
            value = displayData[key][0]
            unit = displayData[key][1]
            if unit:
                print(f"{key}={value}({unit})")
            else:
                print(f"{key}={value}")



""" reading.py """
# from enum import auto
# from strenum import LowercaseStrEnum

from powermon.dto.reading_dto import ReadingDTO


class Reading:
    def __str__(self):
        return f"Reading: {self.data_name=}, {self.data_value=}, {self.data_unit=}, {self.is_valid=}"

    def __init__(self, data_name: str,
                 data_value: str,
                 data_unit: str,
                 device_class: str = None,
                 state_class: str = None,
                 icon: str = None) -> None:
        self.data_name = data_name
        self.data_value = data_value
        self.data_unit = data_unit
        self.device_class = device_class
        self.icon = icon
        self.state_class = state_class
        self.is_valid = True

    def to_dto(self) -> ReadingDTO:
        return ReadingDTO(data_name=self.get_data_name(), data_value=self.get_data_value(), data_unit=self.get_data_unit())

    def get_data_name(self) -> str:
        return self.data_name.replace(" ", "_").lower()

    def get_data_unit(self) -> str:
        if self.data_unit is None:
            return ""
        return self.data_unit

    def get_data_value(self) -> str:
        return self.data_value

    def get_icon(self) -> str | None:
        return self.icon

    def get_device_class(self) -> str | None:
        return self.device_class

    def get_state_class(self) -> str | None:
        return self.state_class



""" outputs / screen.py """
import logging

from powermon.commands.result import Result
# from powermon.device import Device
# from powermon.dto.outputDTO import OutputDTO
# from powermon.formats.abstractformat import AbstractFormat
from powermon.outputs.abstractoutput import AbstractOutput

log = logging.getLogger("screen")


class Screen(AbstractOutput):
    """outputs the results to the screen as per the formatter supplied """
    def __init__(self):
        super().__init__(name="Screen")

    def __str__(self):
        return "outputs.Screen: outputs the results to the screen as per the formatter supplied"

    def process(self, result: Result, mqtt_broker=None):
        log.info("Using output sender: screen")
        log.debug("formatter: %s", self.formatter)

        formatted_data = self.formatter.format(result)
        if formatted_data is None:
            print("Nothing returned from data formatting")
            return

        for line in formatted_data:
            print(line)

        if result.error:
            print("Errors occurred during processing")
            for error in result.error_messages:
                print(error)

    @classmethod
    def from_config(cls, output_config) -> "Screen":
        """If we need to include any config for the Screen output but the processing here"""
        log.debug("config: %s", output_config)
        return cls()



"""device.py"""
import logging

from powermon.commands.command import Command
from powermon.commands.result import Result, ResultType
from powermon.dto.deviceDTO import DeviceDTO
from powermon.errors import ConfigError
from powermon.outputs.abstractoutput import AbstractOutput
from powermon.ports import from_config as port_from_config
from powermon.ports.abstractport import AbstractPort

# Set-up logger
log = logging.getLogger("Device")


class Device:
    """
    A device is a port with a protocol
    also contains the name, model and id of the device
    """

    def __str__(self):
        return f"Device: {self.name}, {self.device_id=}, {self.model=}, {self.manufacturer=}, \
            port: {self.port}, mqtt_broker: {self.mqtt_broker}, commands:{self.commands}"

    @classmethod
    def from_config(cls, config=None):
        """build the object from a config dict"""
        if not config:
            log.warning("No device definition in config. Check configFile argument?")
            return cls(name="unnamed")
        name = config.get("name", "unnamed_device")
        device_id = config.get("id", "1")  # device_id needs to be unique if there are two devices
        model = config.get("model")
        manufacturer = config.get("manufacturer")

        port = port_from_config(config.get("port"))

        # raise error if unable to configure port
        if not port:
            log.error("Invalid port config '%s' found", config)
            raise ConfigError(f"Invalid port config '{config}' found")

        return cls(name=name, device_id=device_id, model=model, manufacturer=manufacturer, port=port)

    def __init__(self, name: str, device_id: str = "", model: str = "", manufacturer: str = "", port: AbstractPort = None):
        self.name = name
        self.device_id = device_id
        self.model = model
        self.manufacturer = manufacturer
        self.port: AbstractPort = port
        self.commands: list[Command] = []
        self.mqtt_broker = None

    def add_command(self, command: Command) -> None:
        """add a command to the devices' list of commands"""
        log.debug("Adding command: %s", command)
        if command is None:
            return
        # get command definition from protocol
        command_definition = self.port.protocol.get_command_with_command_string(command.code)
        if command_definition is None:
            log.error("Cannot find command code: %s, in protocol: %s", command.code, self.port.protocol.get_protocol_id())
            raise RuntimeError(f"Invalid command code: {command.code}, in protocol: {self.port.protocol.get_protocol_id()}")
        command.set_command_definition(command_definition)
        # set the device_id in the command
        command.set_device_id(self.device_id)
        # append to commands list
        self.commands.append(command)
        log.debug("added command (%s), command list length: %i", command, len(self.commands))

    def get_port(self) -> AbstractPort:
        """return the port associated with this device"""
        return self.port

    def set_mqtt_broker(self, mqtt_broker):
        """ store the mqtt broker """
        self.mqtt_broker = mqtt_broker

    def to_dto(self) -> DeviceDTO:
        """convert the Device to a Data Transfer Object"""
        commands = []
        command: Command
        for command in self.commands:
            commands.append(command.to_dto())
        dto = DeviceDTO(device_id=self.device_id,
                        model=self.model,
                        manufacturer=self.manufacturer,
                        port=self.port.to_dto(),
                        commands=commands)
        return dto

    def initialize(self):
        """Device initialization activities"""
        log.info("initializing device")

    def finalize(self):
        """Device finalization activities"""
        log.info("finalizing device")
        # close connection on port
        self.port.disconnect()

    def run(self, force=False):
        """checks for commands to run and runs them"""
        if self.commands is None or len(self.commands) == 0:
            log.info("no commands in queue")
            return

        for command in self.commands:
            if force or command.is_due():
                log.debug("Running command: %s", command.code)
                try:
                    # run command
                    result: Result = self.port.run_command(command)
                except Exception as exception:  # pylint: disable=W0718
                    log.error("Error decoding result: %s", exception)
                    result = Result(command_code=command.code, result_type=ResultType.ERROR, raw_response=b"Error decoding result")
                    result.error_messages.append(f"Error decoding result: {exception}")
                    result.error_messages.append(f"Exception Type: {exception.__class__.__name__}")
                    result.error_messages.append(f"Exception args: {exception.args}")
                    raise exception
                result.set_device_id(self.device_id)  # FIXME: think this is limiting, should pass device and mqtt_broker to output

                # loop through each output and process result
                output: AbstractOutput
                for output in command.outputs:
                    log.debug("Using Output: %s", output)
                    output.process(result=result, mqtt_broker=self.mqtt_broker)
